﻿using System;
using System.IO;
using System.Linq;
using System.Text;
using System.Collections.Generic;
using Antlr4.Runtime;
using Antlr4.Runtime.Tree;
using CutsceneGrammar;

namespace CutsceneCompiler
{
    public class Compiler
    {
        #region Result
        public class Result
        {
            public Executable Executable { get; internal set; }

            internal List<string> mErrors;
            public IEnumerable<string> Errors { get { return mErrors; } }
            public bool HasErrors { get { return mErrors.Count > 0; } }

            internal Result()
            {
                mErrors = new List<string>();
            }

            internal void AddError(string errorType, string error)
            {
                mErrors.Add(errorType + " error: " + error);
            }
        }
        #endregion

        #region CompileTask
        internal class CompileTask : CutsceneBaseListener, IAntlrErrorListener<IToken>
        {
            #region Routine
            internal class Routine
            {
                internal class Jump
                {
                    public string mRoutineName;
                    public ushort mOffset;

                    public Jump(string routineName, ushort offset)
                    {
                        mRoutineName = routineName;
                        mOffset = offset;
                    }
                }

                public ExecutableWriter mWriter;

                public ushort mBaseOffset;
                public List<Jump> mJumps;

                public string Name { get { return mWriter.Name; } }

                internal Routine(string name)
                {
                    mWriter = new ExecutableWriter(name);

                    mBaseOffset = 0;
                    mJumps = new List<Jump>();
                }

                public bool IsStartingRoutine(string routine)
                {
                    foreach (Jump jmp in mJumps)
                    {
                        if (jmp.mRoutineName == routine)
                            return true;
                    }
                    return false;
                }

                public override string ToString()
                {
                    return Name;
                }
            }
            #endregion

            private Result mResult;

            private List<Routine> mRoutines;
            private Routine MainRoutine { get { return mRoutines.Count > 0 ? mRoutines[0] : null; } }
            private Dictionary<string, Routine> mRoutineMap;
            private Routine mCurrentRoutine;

            private HashSet<string> mCreatedActors;
            private HashSet<string> mUsedActors;

            internal CompileTask(Result result)
            {
                mResult = result;

                mRoutines = new List<Routine>();
                mRoutineMap = new Dictionary<string, Routine>();

                mCurrentRoutine = new Routine("main");
                mRoutines.Add(mCurrentRoutine);

                mCurrentRoutine = null;

                mCreatedActors = new HashSet<string>();
                mUsedActors = new HashSet<string>();
            }

            #region Actor tracking
            private void CreateActor(string actorId)
            {
                if (mCreatedActors.Contains(actorId))
                {
                    mResult.AddError("Sematic", "Actor \"" + actorId + "\" was created multiple times.");
                    return;
                }

                mCreatedActors.Add(actorId);
            }

            private void UseActor(string actorId)
            {
                if (!mUsedActors.Contains(actorId))
                    mUsedActors.Add(actorId);
            }
            #endregion

            #region Handle routines
            public override void EnterRoutine(CutsceneParser.RoutineContext context)
            {
                // Set the currently active routine
                string routineName = context.children[1].ToString();

                if (routineName == "main" && !mRoutineMap.ContainsKey("main"))
                {
                    mCurrentRoutine = MainRoutine;
                }
                else
                {
                    mCurrentRoutine = new Routine(routineName);
                    mRoutines.Add(mCurrentRoutine);
                }

                mRoutineMap[routineName] = mCurrentRoutine;

                // Write the name of the routine
                int nameIdx = MainRoutine.mWriter.AddString(routineName);
                mCurrentRoutine.mWriter.AddLiteral((ushort)nameIdx);
            }

            public override void ExitRoutine(CutsceneParser.RoutineContext context)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.RETURN);
                mCurrentRoutine = null;
            }
            #endregion

            #region Handle functions
            #region Helpers
            private bool ValidateLiteral(ref int val)
            {
                if (val < 0 || val > 0xffff)
                {
                    val = Math.Min(Math.Max(val, 0), 0xffff);
                    return false;
                }

                return true;
            }

            private void Handle_Duration(int seconds, int milliseconds)
            {
                int totalNumMilliseconds = seconds * 1000 + milliseconds;
                if (!ValidateLiteral(ref totalNumMilliseconds))
                    mResult.AddError("Semantic", "Duration too long: " + seconds.ToString() + "s " + milliseconds.ToString() + "ms");

                mCurrentRoutine.mWriter.AddDuration((ushort)totalNumMilliseconds);
            }
            #endregion

            #region StartRoutine
            public override void EnterStatement_StartRoutine(CutsceneParser.Statement_StartRoutineContext context)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.StartRoutine);
                
                string routineId = context.children[1].ToString();
                mCurrentRoutine.mJumps.Add(new Routine.Jump(routineId, mCurrentRoutine.mWriter.DataLength));

                mCurrentRoutine.mWriter.AddAddress(0xdead);

            }
            #endregion

            #region Wait
            private void Handle_Wait(int seconds, int milliseconds)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.Wait);
                Handle_Duration(seconds, milliseconds);
            }

            public override void EnterStatement_Wait(CutsceneParser.Statement_WaitContext context)
            {
                string durationMs = context.children[1].ToString();
                Handle_Wait(0, int.Parse(durationMs));
            }

            public override void  EnterStatement_WaitSec(CutsceneParser.Statement_WaitSecContext context)
            {
                string durationSec = context.children[1].ToString();
                string durationMs = context.children[2].ToString();
                Handle_Wait(int.Parse(durationSec), int.Parse(durationMs));
            }
            #endregion

            #region CreateActor
            private void Handle_CreateActor(string actorId, string actorType)
            {
                CreateActor(actorId);

                mCurrentRoutine.mWriter.AddOpCode(OpCode.CreateActor);

                int actorIdIdx = MainRoutine.mWriter.AddString(actorId);
                mCurrentRoutine.mWriter.AddLiteral((ushort)actorIdIdx);

                actorType = actorType.Trim('"');
                int actorTypeIdx = MainRoutine.mWriter.AddString(actorType);
                mCurrentRoutine.mWriter.AddLiteral((ushort)actorTypeIdx);
            }

            public override void EnterStatement_CreateActor(CutsceneParser.Statement_CreateActorContext context)
            {
                string actorId = context.children[1].ToString();
                string actorType = context.children[2].ToString();
                Handle_CreateActor(actorId, actorType);
            }

            public override void EnterStatement_CreateActorMoveActor(CutsceneParser.Statement_CreateActorMoveActorContext context)
            {
                string actorId = context.children[1].ToString();
                string actorType = context.children[2].ToString();
                Handle_CreateActor(actorId, actorType);

                string x = context.children[3].ToString();
                string y = context.children[4].ToString();
                Handle_MoveActor(actorId, int.Parse(x), int.Parse(y));
            }
            #endregion

            #region MoveActor
            private void Handle_MoveActor(string actorId, int _x, int _y)
            {
                UseActor(actorId);

                mCurrentRoutine.mWriter.AddOpCode(OpCode.MoveActor);

                int actorIdIdx = MainRoutine.mWriter.AddString(actorId);
                mCurrentRoutine.mWriter.AddLiteral((ushort)actorIdIdx);

                int x = _x;
                if (!ValidateLiteral(ref x))
                    mResult.AddError("Semantic", "Numeric overflow: x = " + _x.ToString());
                int y = _y;
                if (!ValidateLiteral(ref y))
                    mResult.AddError("Semantic", "Numeric overflow: y = " + x.ToString());

                mCurrentRoutine.mWriter.AddLiteral((ushort)x);
                mCurrentRoutine.mWriter.AddLiteral((ushort)y);
            }

            public override void EnterStatement_MoveActor(CutsceneParser.Statement_MoveActorContext context)
            {
                string actorId = context.children[1].ToString();
                string x = context.children[2].ToString();
                string y = context.children[3].ToString();
                Handle_MoveActor(actorId, int.Parse(x), int.Parse(y));
            }
            #endregion

            #region WalkActor
            private void Handle_WalkActor(string actorId, int _x, int _y)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.WalkActor);

                int actorIdIdx = MainRoutine.mWriter.AddString(actorId);
                mCurrentRoutine.mWriter.AddLiteral((ushort)actorIdIdx);

                int x = _x;
                if (!ValidateLiteral(ref x))
                    mResult.AddError("Semantic", "Numeric overflow: x = " + _x.ToString());
                mCurrentRoutine.mWriter.AddLiteral((ushort)x);

                int y = _y;
                if (!ValidateLiteral(ref y))
                    mResult.AddError("Semantic", "Numeric overflow: y = " + _y.ToString());
                mCurrentRoutine.mWriter.AddLiteral((ushort)y);

                UseActor(actorId);
            }

            public override void EnterStatement_WalkActor(CutsceneParser.Statement_WalkActorContext context)
            {
                string actorId = context.children[1].ToString();
                string x = context.children[2].ToString();
                Handle_WalkActor(actorId, int.Parse(x), 0);
            }

            public override void EnterStatement_WalkActorXY(CutsceneParser.Statement_WalkActorXYContext context)
            {
                string actorId = context.children[1].ToString();
                string x = context.children[2].ToString();
                string y = context.children[3].ToString();
                Handle_WalkActor(actorId, int.Parse(x), int.Parse(y));
            }
            #endregion

            #region StartAnimActor
            public override void EnterStatement_StartAnimActor(CutsceneParser.Statement_StartAnimActorContext context)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.StartAnimActor);

                string actorId = context.children[1].ToString();
                int actorIdIdx = MainRoutine.mWriter.AddString(actorId);
                mCurrentRoutine.mWriter.AddLiteral((ushort)actorIdIdx);

                string anim = context.children[2].ToString();
                int animIdx = MainRoutine.mWriter.AddString(anim);
                mCurrentRoutine.mWriter.AddLiteral((ushort)animIdx);

                UseActor(actorId);
            }
            #endregion

            #region ShowActorText
            private void Handle_ShowActorText(string actorId, string text, int seconds, int milliseconds)
            {
                UseActor(actorId);

                mCurrentRoutine.mWriter.AddOpCode(OpCode.ShowActorText);

                int actorIdIdx = MainRoutine.mWriter.AddString(actorId);
                mCurrentRoutine.mWriter.AddLiteral((ushort)actorIdIdx);

                string txt = text.Trim('"');
                int textIdx = MainRoutine.mWriter.AddString(txt);
                mCurrentRoutine.mWriter.AddLiteral((ushort)textIdx);

                Handle_Duration(seconds, milliseconds);
            }

            public override void EnterStatement_ShowActorText(CutsceneParser.Statement_ShowActorTextContext context)
            {
                string actorId = context.children[1].ToString();
                string text = context.children[2].ToString();
                string durationMs = context.children[3].ToString();
                Handle_ShowActorText(actorId, text, 0, int.Parse(durationMs));
            }

            public override void EnterStatement_ShowActorTextSec(CutsceneParser.Statement_ShowActorTextSecContext context)
            {
                string actorId = context.children[1].ToString();
                string text = context.children[2].ToString();
                string durationSec = context.children[3].ToString();
                string durationMs = context.children[4].ToString();
                Handle_ShowActorText(actorId, text, int.Parse(durationSec), int.Parse(durationMs));
            }
            #endregion

            #region ShowText
            private void Handle_ShowText(string text, int seconds, int milliseconds)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.ShowText);

                string txt = text.Trim('"');
                int textIdx = MainRoutine.mWriter.AddString(txt);
                mCurrentRoutine.mWriter.AddLiteral((ushort)textIdx);

                Handle_Duration(seconds, milliseconds);
            }

            public override void EnterStatement_ShowText(CutsceneParser.Statement_ShowTextContext context)
            {
                string text = context.children[1].ToString();
                string durationMs = context.children[2].ToString();
                Handle_ShowText(text, 0, int.Parse(durationMs));
            }

            public override void EnterStatement_ShowTextSec(CutsceneParser.Statement_ShowTextSecContext context)
            {
                string text = context.children[1].ToString();
                string durationSec = context.children[2].ToString();
                string durationMs = context.children[3].ToString();
                Handle_ShowText(text, int.Parse(durationSec), int.Parse(durationMs));
            }
            #endregion

            #region Yield
            public override void EnterStatement_Yield(CutsceneParser.Statement_YieldContext context)
            {
                mCurrentRoutine.mWriter.AddOpCode(OpCode.YIELD);
            }
            #endregion
            #endregion

            #region Linker validation
            private bool CheckRoutines()
            {
                bool success = true;

                HashSet<string> routineUniqueness = new HashSet<string>();
                foreach (Routine routine in mRoutines)
                {
                    if (routineUniqueness.Contains(routine.Name))
                    {
                        success = false;
                        mResult.AddError("Syntax", "Routine \"" + routine.Name + "\" was defined multiple times.");
                    }
                    else
                        routineUniqueness.Add(routine.Name);
                }

                if (!mRoutineMap.ContainsKey("main"))
                {
                    success = false;
                    mResult.AddError("Syntax", "Routine \"main\" was not defined.");
                }

                return success;
            }

            private bool CheckActors()
            {
                bool success = true;

                foreach (string actorId in mUsedActors)
                {
                    if (!mCreatedActors.Contains(actorId))
                    {
                        mResult.AddError("Link", "Actor \"" + actorId + "\" was used but never created.");
                        success = false;
                    }
                }

                return success;
            }
            #endregion

            public void Link()
            {
                // Check for linker errors
                if (!CheckRoutines())
                    return;
                if (!CheckActors())
                    return;

                // Add the other non-main routines
                for (int i = 1; i < mRoutines.Count; i++)
                {
                    Routine routine = mRoutines[i];

                    // Is the routine started at all?
                    bool isUsed = false;
                    foreach (Routine r in mRoutines)
                    {
                        if (r.IsStartingRoutine(routine.Name))
                        {
                            isUsed = true;
                            break;
                        }
                    }

                    routine.mBaseOffset = 0xffff;
                    if (isUsed)
                    {
                        // Embed the routine (since it's used) and remember where it starts
                        routine.mBaseOffset = MainRoutine.mWriter.DataLength;
                        MainRoutine.mWriter.AddExecutable(routine.mWriter);
                    }
                }

                // Update the jump offsets
                foreach (Routine routine in mRoutines)
                {
                    if (routine.mBaseOffset < 0xffff)
                    {
                        foreach (Routine.Jump jmp in routine.mJumps)
                        {
                            bool found = false;
                            foreach (Routine r in mRoutines)
                            {
                                if (r != routine)
                                {
                                    if (r.Name == jmp.mRoutineName)
                                    {
                                        ushort jumpOrigin = (ushort)(routine.mBaseOffset + jmp.mOffset);
                                        MainRoutine.mWriter.InsertAddress(jumpOrigin, r.mBaseOffset);

                                        found = true;
                                        break;
                                    }
                                }
                            }

                            if (!found)
                                mResult.AddError("Link", "Can't start undefined routine \"" + jmp.mRoutineName + "\".");
                        }
                    }
                }

                // Everything looks good, so let's create the executable
                mResult.Executable = MainRoutine.mWriter.CreateExecutable();
            }

            public void SyntaxError(IRecognizer recognizer, IToken offendingSymbol, int line, int charPositionInLine, string msg, RecognitionException e)
            {
                mResult.AddError("Syntax", msg + " (line " + line.ToString() + " column " + charPositionInLine.ToString() + ")");
            }
        }
        #endregion

        private ParseTreeWalker mParsedCutsceneTreeWalker;

        public Compiler()
        {
            mParsedCutsceneTreeWalker = new ParseTreeWalker();
        }

        public Result Compile(string cutsceneSourceCode)
        {
            Result result = new Result();

            // Pass the source code to the lexer...
            AntlrInputStream inputStream = new AntlrInputStream(cutsceneSourceCode);
            CutsceneLexer lexer = new CutsceneLexer(inputStream);
            CommonTokenStream tokenStream = new CommonTokenStream(lexer);
            // ...before setting up the parser
            CutsceneParser parser = new CutsceneParser(tokenStream);
            CompileTask task = new CompileTask(result);
            parser.AddErrorListener(task);
            // Parse the given cutscene
            IParseTree parsedCutscene = parser.cutscene();
            if (parser.NumberOfSyntaxErrors == 0)
            {
                mParsedCutsceneTreeWalker.Walk(task, parsedCutscene);

                // Create the final executable
                task.Link();
            }

            return result;
        }
    }
}
