﻿using System;
using System.Text;

namespace CutsceneCompiler
{
    public static class Disassembler
    {
        public static string Disassemble(Executable executable)
        {
            bool success = true;
            StringBuilder cutscene = new StringBuilder();

            ushort offset = 0;
            while (offset < executable.DataSize)
            {
                if (!DisassembleRoutine(executable, ref offset, cutscene))
                {
                    success = false;
                    break;
                }
            }

            success &= offset > 0;

            return success ? cutscene.ToString() : null;
        }

        private static bool DisassembleRoutine(Executable executable, ref ushort offset, StringBuilder cutscene)
        {
            bool success = false;

            if (offset < executable.DataSize)
            {
                string routineName = DisassembleString(executable, ref offset);
                if (!string.IsNullOrEmpty(routineName))
                {
                    cutscene.AppendLine("routine " + routineName + " {");
                    success = DisassembleRoutineBody(executable, ref offset, cutscene);
                    cutscene.AppendLine("}");
                    cutscene.AppendLine();
                }
            }

            return success;
        }

        private static bool DisassembleRoutineBody(Executable executable, ref ushort offset, StringBuilder cutscene)
        {
            bool success = true;

            ushort numOpCodes = (ushort)Enum.GetNames(typeof(OpCode)).Length;

            bool isDone = false;
            while (!isDone)
            {
                bool error = true;

                if (offset < executable.DataSize)
                {
                    ushort opCode = executable[offset++];
                    if (opCode < numOpCodes && opCode != (ushort)OpCode.SETPC)
                    {
                        if (opCode == (ushort)OpCode.RETURN)
                        {
                            isDone = true;
                            error = false;
                        }
                        else
                            error = !DisassembleOperation((OpCode)opCode, executable, ref offset, cutscene);
                    }
                }

                if (error)
                {
                    success = false;
                    isDone = true;
                }
            }


            return success;
        }

        private static bool DisassembleOperation(OpCode opCode, Executable executable, ref ushort offset, StringBuilder cutscene)
        {
            bool success = false;

            string statement = null;
            string opCodeString = opCode.ToString();
            switch (opCode)
            {
                case OpCode.YIELD:
                    statement = "Yield";
                    break;

                case OpCode.StartRoutine:
                    {
                        string routineName = DisassembleRoutineName(executable, ref offset);
                        if (!string.IsNullOrEmpty(routineName))
                            statement = opCodeString + " " + routineName;

                    } break;

                case OpCode.Wait:
                    {
                        string duration = DisassembleDuration(executable, ref offset);
                        if (!string.IsNullOrEmpty(duration))
                            statement = opCodeString + " " + duration;

                    } break;

                case OpCode.CreateActor:
                    {
                        string actorId = DisassembleString(executable, ref offset);
                        string actorType = DisassembleString(executable, ref offset);
                        if (!string.IsNullOrEmpty(actorId) && !string.IsNullOrEmpty(actorType))
                            statement = opCodeString + " " + actorId + " \"" + actorType + "\"";

                    } break;

                case OpCode.MoveActor:
                case OpCode.WalkActor:
                    {
                        string actorId = DisassembleString(executable, ref offset);
                        ushort? x = DisassembleLiteral(executable, ref offset);
                        ushort? y = DisassembleLiteral(executable, ref offset);
                        if (!string.IsNullOrEmpty(actorId) && x.HasValue && y.HasValue)
                            statement = opCodeString + " " + actorId + " " + x.Value.ToString() + " " + y.Value.ToString();

                    } break;

                case OpCode.StartAnimActor:
                    {
                        string actorId = DisassembleString(executable, ref offset);
                        string actorType = DisassembleString(executable, ref offset);
                        if (!string.IsNullOrEmpty(actorId) && !string.IsNullOrEmpty(actorType))
                            statement = opCodeString + " " + actorId + " " + actorType;

                    } break;

                case OpCode.ShowActorText:
                    {
                        string actorId = DisassembleString(executable, ref offset);
                        string text = DisassembleString(executable, ref offset);
                        string duration = DisassembleDuration(executable, ref offset);
                        if (!string.IsNullOrEmpty(actorId) && !string.IsNullOrEmpty(text) && !string.IsNullOrEmpty(duration))
                            statement = opCodeString + " " + actorId + " \"" + text + "\"" + " " + duration;

                    } break;

                case OpCode.ShowText:
                    {
                        string text = DisassembleString(executable, ref offset);
                        string duration = DisassembleDuration(executable, ref offset);
                        if (!string.IsNullOrEmpty(text) && !string.IsNullOrEmpty(duration))
                            statement = opCodeString + " \"" + text + "\"" + " " + duration;

                    } break;
            }

            if (!string.IsNullOrEmpty(statement))
            {
                success = true;
                cutscene.AppendLine("\t" + statement + ";");
            }

            return success;
        }

        private static string DisassembleString(Executable executable, ref ushort offset)
        {
            ushort? stringIdx = DisassembleLiteral(executable, ref offset);
            return stringIdx.HasValue && stringIdx.Value < executable.StringTableSize ? executable.GetString(stringIdx.Value) : null;
        }

        private static string DisassembleRoutineName(Executable executable, ref ushort offset)
        {
            if (offset + 1 >= executable.DataSize)
                return null;

            int high = executable[offset++];
            int low = executable[offset++];
            ushort address = (ushort)((high << 8) + low);

            if (address >= executable.DataSize)
                return null;

            ushort stringIdx = executable[address++];
            return stringIdx < executable.StringTableSize ? executable.GetString(stringIdx) : null;
        }

        private static string DisassembleDuration(Executable executable, ref ushort offset)
        {
            string duration = null;

            ushort? milliseconds = DisassembleLiteral(executable, ref offset);
            if (milliseconds.HasValue)
            {
                if (milliseconds.Value > 1000)
                {
                    int seconds = milliseconds.Value / 1000;
                    int millisecondsLeft = milliseconds.Value - (seconds * 1000);
                    duration = seconds.ToString() + " " + millisecondsLeft.ToString();
                }
                else
                {
                    duration = milliseconds.Value.ToString();
                }
            }

            return duration;
        }

        private static ushort? DisassembleLiteral(Executable executable, ref ushort offset)
        {
            if (offset >= executable.DataSize)
                return null;

            ushort value = executable[offset++];
            if ((value & 0x0080) != 0)
            {
                if ((value & 0x00c0) == 0x0080)
                {
                    if (offset >= executable.DataSize)
                        return null;

                    int high = value & 0x003f;
                    int low = executable[offset++];
                    value = (ushort)((high << 8) + low);
                }
                else
                {
                    if ((offset + 1) >= executable.DataSize)
                        return null;

                    int high = executable[offset++];
                    int low = executable[offset++];
                    value = (ushort)((high << 8) + low);
                }
            }

            return value;
        }
    }
}
