﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace CutsceneEditor
{
    public class Actor
    {
        public static readonly int UnitsPerCell = 16;

        public string Name { get; private set; }

        public Color Color { get; private set; }

        private Dictionary<string, Animation> mAnimations;

        private string mNameCurAnimation;
        public Animation CurrentAnimation 
        { 
            get
            {
                Animation animation;
                if (!string.IsNullOrEmpty(mNameCurAnimation) && mAnimations.TryGetValue(mNameCurAnimation, out animation))
                    return animation;

                return null;
            } 
        }

        public int X { get; set; }
        public int Y { get; set; }

        public int U { get; set; }
        public int V { get; set; }

        public int DU { get; set; }
        public int DV { get; set; }

        public Actor(string name)
        {
            Name = name;

            Color = Color.Black;

            mAnimations = new Dictionary<string, Animation>();

            X = 0;
            Y = 0;

            U = 0;
            V = 0;

            DU = 0;
            DV = 0;
        }

        private void AddAnimation(Animation animation)
        {
            mAnimations.Add(animation.Name, animation);

            if (string.IsNullOrEmpty(mNameCurAnimation))
                StartAnimation(animation.Name);
        }

        public void StartAnimation(string name)
        {
            if (CurrentAnimation != null)
                CurrentAnimation.Stop();

            Animation animation;
            if (mAnimations.TryGetValue(name, out animation))
            {
                mNameCurAnimation = name;
                CurrentAnimation.Start();
            }
        }

        public void OnTick()
        {
            U += DU;
            X += U / UnitsPerCell;
            U = U % UnitsPerCell;

            V += DV;
            Y += V / UnitsPerCell;
            V = V % UnitsPerCell;

            CurrentAnimation.OnTick();
        }

        public override string ToString()
        {
            return Name;
        }

        public static Actor CreateActor(string name, string type)
        {
            Actor result = null;

            ActorDefinition actorDef = ActorDefinition.GetActorDefinition(type);
            if (actorDef != null)
            {
                result = new Actor(name);

                result.Color = Color.FromArgb(actorDef.ColorR, actorDef.ColorG, actorDef.ColorB);

                foreach (ActorDefinition.Animation animDef in actorDef.Animations)
                {
                    ImageSequence animFrames = new ImageSequence();

                    foreach (string animFrameDef in animDef.Frames)
                    {
                        FileInfo animFrameFile = new FileInfo(Path.Combine(actorDef.File.DirectoryName, animFrameDef));
                        if (!animFrames.AddImageFromFile(animFrameFile.FullName))
                        {
                            MessageBox.Show("Unable to open animation frame: " + animFrameFile.Name, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                    }

                    Animation anim = new Animation(animDef.Name, animFrames);
                    anim.IsFlipped = animDef.IsFlipped;
                    result.AddAnimation(anim);
                }
            }

            return result;
        }
    }
}
