﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Windows.Forms;
using Antlr4.Runtime;
using Antlr4.Runtime.Tree;

namespace CutsceneEditor
{
    public class ActorDefinition
    {
        private static DirectoryInfo gActorRootDirectory;
        private static Dictionary<string, ActorDefinition> gActorDefinitions;

        #region Parser
        internal class ParseTask : ActorBaseListener, IAntlrErrorListener<IToken>
        {
            private ActorDefinition mActorDef;
            private Animation mCurrentAnimation;

            public ParseTask(ActorDefinition actorDef)
            {
                mActorDef = actorDef;
                mCurrentAnimation = null;
            }

            public override void EnterPropertyColorR(ActorParser.PropertyColorRContext context)
            {
                try
                {
                    mActorDef.ColorR = Math.Min(Math.Max(int.Parse(context.children[2].ToString()), 0), 255);
                }
                catch
                {
                }
            }

            public override void EnterPropertyColorG(ActorParser.PropertyColorGContext context)
            {
                try
                {
                    mActorDef.ColorG = Math.Min(Math.Max(int.Parse(context.children[2].ToString()), 0), 255);
                }
                catch
                {
                }
            }

            public override void EnterPropertyColorB(ActorParser.PropertyColorBContext context)
            {
                try
                {
                    mActorDef.ColorB = Math.Min(Math.Max(int.Parse(context.children[2].ToString()), 0), 255);
                }
                catch
                {
                }
            }

            public override void EnterAnimation(ActorParser.AnimationContext context)
            {
                string animationName = context.children[0].ToString();
                mCurrentAnimation = new Animation(animationName);
            }

            public override void ExitAnimation(ActorParser.AnimationContext context)
            {
                mActorDef.Animations.Add(mCurrentAnimation);
                mCurrentAnimation = null;
            }

            public override void EnterFlippedAnimation(ActorParser.FlippedAnimationContext context)
            {
                mCurrentAnimation.IsFlipped = true;
            }

            public override void EnterFrame(ActorParser.FrameContext context)
            {
                string frameFilename = context.children[0].ToString();
                mCurrentAnimation.Frames.Add(frameFilename);
            }

            public void SyntaxError(IRecognizer recognizer, IToken offendingSymbol, int line, int charPositionInLine, string msg, RecognitionException e)
            {
                string error = msg + " (line " + line.ToString() + " column " + charPositionInLine.ToString() + ")";
                MessageBox.Show(error, "Syntax error in actor: " + mActorDef.File.Name, MessageBoxButtons.OK, MessageBoxIcon.Error);
                
            }
        }
        #endregion

        #region Animation
        public class Animation
        {
            public string Name { get; private set; }
            public bool IsFlipped { get; set; }
            public List<string> Frames { get; private set; }

            public Animation(string name)
            {
                Name = name;
                IsFlipped = false;
                Frames = new List<string>();
            }
        }
        #endregion

        public FileInfo File { get; private set; }
        public int ColorR { get; set; }
        public int ColorG { get; set; }
        public int ColorB { get; set; }
        public List<Animation> Animations { get; private set; }

        private ActorDefinition()
        {
            File = null;
            Animations = new List<Animation>();
        }

        #region Deserialization
        private bool FromFile(string filename)
        {
            File = new FileInfo(filename);
            if (!File.Exists)
            {
                MessageBox.Show("Unable to open actor file: " + File.Name, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            StreamReader reader = File.OpenText();
            string actorDef = reader.ReadToEnd();
            reader.Close();

            return ParseActor(actorDef);
        }

        private bool ParseActor(string actorDef)
        {
            bool success = false;

            // Pass the actor defintion to the lexer...
            AntlrInputStream inputStream = new AntlrInputStream(actorDef);
            ActorLexer lexer = new ActorLexer(inputStream);
            CommonTokenStream tokenStream = new CommonTokenStream(lexer);
            // ...before setting up the parser
            ActorParser parser = new ActorParser(tokenStream);
            ParseTask task = new ParseTask(this);
            parser.AddErrorListener(task);
            // Parse the given actor
            IParseTree parsedActor = parser.actor();
            if (parser.NumberOfSyntaxErrors == 0)
            {
                ParseTreeWalker parsedActorTreeWalker = new ParseTreeWalker();
                parsedActorTreeWalker.Walk(task, parsedActor);
                success = true;
            }

            return success;
        }
        #endregion

        public static ActorDefinition GetActorDefinition(string name)
        {
            ActorDefinition def;
            if (!gActorDefinitions.TryGetValue(name, out def))
            {
                string actorFilename = Path.Combine(gActorRootDirectory.FullName, name) + ".actr";
                def = new ActorDefinition();
                if (def.FromFile(actorFilename))
                {
                    gActorDefinitions.Add(name, def);
                }
                else
                {
                    def = null;
                }
            }

            return def;
        }

        public static bool Init(string actorRootDirectory)
        {
            gActorRootDirectory = new DirectoryInfo(actorRootDirectory);
            gActorDefinitions = new Dictionary<string, ActorDefinition>();
            return gActorRootDirectory.Exists;
        }
    }
}
