﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Text;
using System.IO;
using System.Reflection;
using CutsceneCompiler;

namespace CutsceneEditor
{
    public partial class CutsceneEditor : Form
    {
        #region Imports
        private const int EM_SETTABSTOPS = 0x00CB;

        [DllImport("User32.dll", CharSet = CharSet.Auto)]
        private static extern IntPtr SendMessage(IntPtr h, int msg, int wParam, int[] lParam);
        #endregion

        private static readonly int kNumVerticalCells = 2;
        private static readonly int kNumHorizontalCells = 16;

        private Timer mClock;

        private Timer mCompileCooldown;
        private bool mIsCompiling;

        private Font mTextFont;

        private DirectoryInfo mRootDirectory;
        private FileInfo mCutsceneFile;

        private Compiler mCompiler;

        private Scene mScene;
        private Script mScript;

        public CutsceneEditor()
        {
            InitializeComponent();

            // Create infrastructure stuff
            mClock = new Timer();
            mClock.Interval = 100;
            mClock.Tick += new EventHandler(mClock_Tick);

            mCompileCooldown = new Timer();
            mCompileCooldown.Interval = 400;
            mCompileCooldown.Tick += new EventHandler(mCompileCooldown_Tick);
            mIsCompiling = false;

            mTextFont = new Font(Font.FontFamily, 18, FontStyle.Bold);

            // Get root directory
            Assembly executable = Assembly.GetExecutingAssembly();
            mRootDirectory = new DirectoryInfo(Path.GetDirectoryName(executable.Location));

            // Init GUI state
            mControlMenuPause.Enabled = false;
            mControlMenuStop.Enabled = false;

            UpdatePresets();
            // ToDo: Use FileSystemWatcher to keep preset list updated

            // Setup event handlers
            mPicBoxCutscene.Paint += new PaintEventHandler(mPicBoxCutscene_Paint);
            mMainMenuFile.DropDownOpening += new EventHandler(mMainMenuFile_DropDownOpening);
            mMainMenuPresets.SelectedIndexChanged += new EventHandler(mMainMenuPresets_SelectedIndexChanged);

            // Setup the cutscene compiler
            mCompiler = new Compiler();

            // Init actor system and scene
            ActorDefinition.Init(Path.Combine(mRootDirectory.FullName, "Actors"));

            mScene = new Scene();
            mScript = Script.CreateTestScript(mScene, 1);
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            // Setup and activate the source code editor
            SendMessage(mTxtBoxEditor.Handle, EM_SETTABSTOPS, 1, new int[] { 16 });
            ActiveControl = mTxtBoxEditor;

            // Set caption
            SetCutsceneFile(null);

            // Set the initial source code
            SetPreset("Syntax");

            mClock.Start();
        }

        #region Tick and draw
        private void mClock_Tick(object sender, EventArgs e)
        {
            mClock.Stop();

            // Update the actors in the scene
            if (mScript.IsRunning)
            {
                foreach (Actor actor in mScene.Actors)
                    actor.OnTick();
            }

            // Redraw the scene
            mPicBoxCutscene.Invalidate();

            // Update GUI state
            UpdatePlaybackControl();

            mClock.Start();
        }

        private void DrawScene(Graphics graph, Rectangle rectView, int cellSize)
        {
            int unitSize = cellSize / Actor.UnitsPerCell;

            foreach (Actor actor in mScene.Actors)
            {
                // Calculate the bounds of the actor
                int x0 = rectView.Left + actor.X * cellSize + actor.U * unitSize;
                int x1 = x0 + cellSize;

                int y1 = rectView.Bottom - actor.Y * cellSize - actor.V * unitSize;
                int y0 = y1 - cellSize;

                if (actor.CurrentAnimation.IsFlipped)
                {
                    x0 = x1;
                    x1 -= cellSize;
                }

                // Draw the current frame of the actor
                Image curFrame = actor.CurrentAnimation.CurrentFrame;
                graph.DrawImage(curFrame, x0, y0, x1 - x0, y1 - y0);
            }

            // Compute the required space for the visible text
            int textHeight = 0;
            foreach (Text text in mScene.Texts)
            {
                SizeF textSize = graph.MeasureString(text.ToString(), mTextFont);
                textHeight += (int)textSize.Height;
            }

            if (textHeight > 0)
            {
                // Draw the visible text
                int y = rectView.Bottom - cellSize - textHeight;

                foreach (Text text in mScene.Texts)
                {
                    SizeF textSize = graph.MeasureString(text.Value, mTextFont);

                    int xCenter = rectView.Width / 2;
                    int yOffset = 0;
                    Brush brush = Brushes.Black;

                    if (text.Actor != null)
                    {
                        xCenter = text.Actor.X * cellSize + text.Actor.U * unitSize + cellSize / 2;
                        yOffset = (text.Actor.Name.GetHashCode() % 2) == 0 ? 0 : -cellSize / 2;
                        brush = new SolidBrush(text.Actor.Color);
                    }

                    int x = xCenter - (int)(textSize.Width * 0.5f);

                    graph.DrawString(text.Value, mTextFont, brush, (float)x, (float)(y + yOffset));

                    y += (int)textSize.Height;
                }
            }
        }
        #endregion

        #region Paint callback
        private void mPicBoxCutscene_Paint(object sender, PaintEventArgs e)
        {
            Graphics graph = e.Graphics;

            int cellSize = 0;
            Rectangle rectView = CalculateViewport(ref cellSize);

            Pen penFrame = Pens.DarkBlue;
            graph.DrawRectangle(penFrame, rectView);

            int x = rectView.Left;
            graph.DrawLine(penFrame, x, rectView.Bottom, x, rectView.Bottom + 10);
            for (int i = 0; i < kNumHorizontalCells; i++)
            {
                int x1 = Math.Min(rectView.Left + cellSize * (i + 1), rectView.Right);
                graph.DrawLine(penFrame, x1, rectView.Bottom, x1, rectView.Bottom + 10);

                int xc = (x + x1) / 2;
                string sIdx = i.ToString();
                SizeF sIdxSize = graph.MeasureString(sIdx, Font);
                graph.DrawString(sIdx, Font, Brushes.DarkBlue, xc - sIdxSize.Width / 2, rectView.Bottom + 10);

                x = x1;
            }

            Region orgClip = graph.Clip;
            graph.Clip = new Region(rectView);

            DrawScene(graph, rectView, cellSize);

            // Draw the run time
            uint runTimeMs = mScript.RunTimeMs;

            uint runTimeMin = runTimeMs / (1000 * 60);
            runTimeMs -= runTimeMin * (1000 * 60);

            uint runTimeSec = runTimeMs / 1000;
            runTimeMs -= runTimeSec * 1000;

            string duration = runTimeMin.ToString("00") + ":" + runTimeSec.ToString("00") + ":" + runTimeMs.ToString("0000");

            graph.Clip = orgClip;
            graph.DrawString(duration, Font, Brushes.Black, ClientSize.Width - 80, 10);
        }

        private Rectangle CalculateViewport(ref int cellSize)
        {
            int margin = 25;
            int marginTop = 25;
            int marginBottom = 50;

            int viewY0 = marginTop;
            int viewY1 = mPicBoxCutscene.ClientSize.Height - marginBottom;
            int viewHeight = viewY1 - viewY0;

            cellSize = viewHeight / kNumVerticalCells;
            int viewWidth = cellSize * kNumHorizontalCells;

            int viewX0 = mPicBoxCutscene.ClientSize.Width / 2 - viewWidth / 2;
            int viewX1 = viewX0 + viewWidth;

            if (mPicBoxCutscene.ClientSize.Width < viewWidth)
            {
                viewWidth = mPicBoxCutscene.ClientSize.Width - margin * 2;
                viewX0 = margin;
                viewX1 = viewX0 + viewWidth;

                cellSize = viewWidth / kNumHorizontalCells;
                viewHeight = cellSize * kNumVerticalCells;

                viewY0 = viewY1 - viewHeight;
            }

            return new Rectangle(viewX0, viewY0, viewWidth, viewHeight);
        }
        #endregion

        #region Compiling
        private Compiler.Result CompileCutscene()
        {
            return mCompiler.Compile(mTxtBoxEditor.Text);
        }

        private bool IsCutsceneCompiling()
        {
            Compiler.Result result = CompileCutscene();
            return !result.HasErrors;
        }
        #endregion

        #region Playback control
        private void mControlMenuStart_Click(object sender, EventArgs e)
        {
            if (!mScript.IsPaused)
            {
                // Compile the current cutscene and replace the script if it's valid
                Compiler.Result result = CompileCutscene();
                mScript = result.HasErrors ? mScript : Script.CreateScript(mScene, result.Executable);
            }

            mScript.Start(false);
        }

        private void mControlMenuPause_Click(object sender, EventArgs e)
        {
            mScript.Pause();
        }

        private void mControlMenuStop_Click(object sender, EventArgs e)
        {
            mScript.Stop();
        }

        private void UpdatePlaybackControl()
        {
            if (mScript.IsRunning)
            {
                mControlMenuStart.Enabled = false;
                mControlMenuPause.Enabled = true;
                mControlMenuStop.Enabled = true;
            }
            else
            {
                mControlMenuStart.Enabled = true;
                mControlMenuPause.Enabled = false;
                mControlMenuStop.Enabled = false;
            }

            foreach (Routine activeRoutine in mScript.Routines)
            {
                if (!mListBoxRoutines.Items.Contains(activeRoutine.Name))
                    mListBoxRoutines.Items.Add(activeRoutine.Name);
            }

            for (int i = mListBoxRoutines.Items.Count - 1; i >= 0; i--)
            {
                string activeRoutineName = mListBoxRoutines.Items[i].ToString();
                if (mScript.GetRoutine(activeRoutineName) == null)
                    mListBoxRoutines.Items.RemoveAt(i);
            }
        }
        #endregion

        #region Editing
        private void mTxtBoxEditor_TextChanged(object sender, EventArgs e)
        {
            string presetCode = GetPresetCode(GetCurrentPreset());
            if (presetCode != mTxtBoxEditor.Text)
                mMainMenuPresets.SelectedIndex = 0;

            mCompileCooldown.Stop();
            mCompileCooldown.Start();
        }

        private void mCompileCooldown_Tick(object sender, EventArgs e)
        {
            mCompileCooldown.Stop();

            if (!mIsCompiling)
            {
                mIsCompiling = true;

                Compiler.Result result = CompileCutscene();

                mTxtBoxIssues.Clear();
                if (result.HasErrors)
                {
                    mTxtBoxIssues.BackColor = Color.MistyRose;
                    mTxtBoxIssues.Lines = result.Errors.ToArray();
                }
                else
                {
                    mTxtBoxIssues.BackColor = Color.Honeydew;
                    mTxtBoxIssues.Text = "Cutscene compiled successfully.";
                }

                mIsCompiling = false;
            }
        }
        #endregion

        #region Presets
        private string GetPresetCode(string presetName)
        {
            if (presetName.ToLower().Equals("custom"))
                return null;

            FileInfo presetFile = new FileInfo(Path.Combine(Path.Combine(mRootDirectory.FullName, "Presets"), presetName) + ".ctsc");
            if (!presetFile.Exists)
            {
                MessageBox.Show("Unable to open preset: " + presetFile.Name, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }

            StreamReader reader = presetFile.OpenText();
            string presetCode = reader.ReadToEnd();
            reader.Close();

            return presetCode;
        }

        private void UpdatePresets()
        {
            string currentPreset = mMainMenuPresets.SelectedItem != null ? mMainMenuPresets.SelectedItem.ToString() : null;

            mMainMenuPresets.BeginUpdate();

            mMainMenuPresets.Items.Clear();
            mMainMenuPresets.Items.Add("Custom");

            DirectoryInfo presetsFolder = new DirectoryInfo(Path.Combine(mRootDirectory.FullName, "Presets"));
            if (presetsFolder.Exists)
            {
                FileInfo[] aPresetFiles = presetsFolder.GetFiles("*.ctsc");
                foreach (FileInfo presetFile in aPresetFiles)
                {
                    string presetName = Path.GetFileNameWithoutExtension(presetFile.Name);
                    mMainMenuPresets.Items.Add(presetName);

                    if (string.IsNullOrEmpty(currentPreset) && presetName.ToLower().Equals("syntax"))
                    {
                        currentPreset = presetName;
                    }
                }
            }

            mMainMenuPresets.EndUpdate();

            if (string.IsNullOrEmpty(currentPreset))
            {
                currentPreset = "Custom";
            }

            SetSelectedPreset(currentPreset);
        }

        private void SetSelectedPreset(string presetName)
        {
            for (int i = 0; i < mMainMenuPresets.Items.Count; i++)
            {
                if (mMainMenuPresets.Items[i].ToString().Equals(presetName))
                {
                    mMainMenuPresets.SelectedIndex = i;
                    break;
                }
            }
        }

        private void SetPreset(string presetName)
        {
            if (presetName.ToLower().Equals("custom"))
                return;

            SetSelectedPreset(presetName);

            string presetCode = GetPresetCode(presetName);
            if (!string.IsNullOrEmpty(presetCode))
            {
                mTxtBoxEditor.Lines = presetCode.Replace("\r", "").Split('\n');
                mTxtBoxEditor.SelectionStart = 0;
                mTxtBoxEditor.SelectionLength = 0;
            }
        }

        private void mMainMenuPresets_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetPreset(GetCurrentPreset());
        }

        private string GetCurrentPreset()
        {
            return mMainMenuPresets.SelectedItem.ToString();
        }
        #endregion

        #region Render
        private bool Render(string outputFilename, int outputHeight, int outputWidth, int outputRate, int maxDurationSec)
        {
            int cellSize = outputHeight / kNumVerticalCells;
            int renderWidth = cellSize * kNumHorizontalCells;
            Rectangle outputRect = new Rectangle(0, 0, renderWidth, outputHeight);

            int renderFrequency = Math.Max(1, mScript.ClockSpeed / outputRate);
            int actorTickFrequency = Math.Max(1, ((mClock.Interval * mScript.ClockSpeed) / 1000) - 3);

            bool success = false;

            mScript.Stop();

            Compiler.Result result = CompileCutscene();
            if (!result.HasErrors)
            {
                List<Bitmap> aFrames = new List<Bitmap>();

                // Execute the script and render the frames
                mScript = Script.CreateScript(mScene, result.Executable);
                mScript.Start(true);

                int frameCount = 0;
                while (!mScript.IsDone)
                {
                    if (mScript.RunTimeMs >= (maxDurationSec * 1000))
                    {
                        mScript.Stop();
                        break;
                    }

                    mScript.Tick();

                    if ((frameCount % actorTickFrequency) == 0)
                    {
                        // Update the position and animation of all actors
                        foreach (Actor actor in mScene.Actors)
                            actor.OnTick();
                    }

                    if ((frameCount % renderFrequency) == 0)
                    {
                        // Draw the current state into a frame
                        Bitmap frame = new Bitmap(renderWidth, outputHeight, PixelFormat.Format24bppRgb);
                        Graphics graph = Graphics.FromImage(frame);

                        graph.FillRectangle(Brushes.White, outputRect);

                        DrawScene(graph, outputRect, cellSize);

                        if (outputWidth != renderWidth)
                        {
                            Bitmap trimmedFrame = new Bitmap(outputWidth, outputHeight, PixelFormat.Format24bppRgb);
                            graph = Graphics.FromImage(trimmedFrame);

                            int xOffset = (outputWidth - renderWidth) / 2;
                            graph.DrawImageUnscaled(frame, xOffset, 0, outputWidth, outputHeight);

                            frame.Dispose();
                            frame = trimmedFrame;
                        }

                        aFrames.Add(frame);
                    }

                    frameCount++;
                }

                // Write the frames into a GIF
                if (aFrames.Count > 0)
                {
                    success = AnimatedGIF.SaveFromFrames(outputFilename, aFrames, renderFrequency);
                }
            }

            return success;
        }

        private void mMainMenuRender_Click(object sender, EventArgs e)
        {
            RenderOptionsDialog renderOptions = new RenderOptionsDialog();

            string rootDir = Path.GetDirectoryName(renderOptions.Filename);
            string filename = mCutsceneFile != null ? Path.GetFileNameWithoutExtension(mCutsceneFile.Name) : GetCurrentPreset();
            renderOptions.Filename = Path.Combine(rootDir, filename + ".gif");

            if (renderOptions.ShowDialog() == DialogResult.OK)
            {
                Render(renderOptions.Filename, renderOptions.ImageHeight, renderOptions.ImageWidth, renderOptions.FramesPerSecond, renderOptions.MaxDurationSec);
            }
        }
        #endregion

        #region Main menu handlers: File
        private void SetCutsceneFile(string filename)
        {
            mCutsceneFile = null;
            if (!string.IsNullOrEmpty(filename))
            {
                mCutsceneFile = new FileInfo(filename);
                if (!mCutsceneFile.Exists)
                    mCutsceneFile = null;
            }

            Text = "Cutscene Editor";
            if (mCutsceneFile != null)
                Text += " - " + mCutsceneFile.Name;
        }

        private void mMainMenuFile_DropDownOpening(object sender, EventArgs e)
        {
            mMainMenuFileSave.Enabled = mCutsceneFile != null;
            mMainMenuFileSaveExeAs.Enabled = IsCutsceneCompiling();
        }

        private void mMainMenuFileNew_Click(object sender, EventArgs e)
        {
            SetPreset("Empty");
            SetCutsceneFile(null);
        }

        private void mMainMenuFileOpen_Click(object sender, EventArgs e)
        {
            mOfdCutscene.FileName = mCutsceneFile != null ? mCutsceneFile.FullName : "";
            mOfdCutscene.InitialDirectory = !string.IsNullOrEmpty(mOfdCutscene.FileName) ? Path.GetDirectoryName(mOfdCutscene.FileName) : ".";
            if (mOfdCutscene.ShowDialog() == DialogResult.OK)
            {
                SetCutsceneFile(mOfdCutscene.FileName);

                StreamReader reader = mCutsceneFile.OpenText();
                string cutscene = reader.ReadToEnd();
                reader.Close();

                mTxtBoxEditor.Text = cutscene;
            }
        }

        private void mMainMenuFileSave_Click(object sender, EventArgs e)
        {
            if (mCutsceneFile != null)
            {
                StreamWriter writer = mCutsceneFile.CreateText();
                writer.Write(mTxtBoxEditor.Text);
                writer.Close();
            }
        }

        private void mMainMenuFileSaveAs_Click(object sender, EventArgs e)
        {
            mSfdCutscene.FileName = mCutsceneFile != null ? mCutsceneFile.FullName : "";
            mSfdCutscene.InitialDirectory = !string.IsNullOrEmpty(mSfdCutscene.FileName) ? Path.GetDirectoryName(mSfdCutscene.FileName) : ".";
            if (mSfdCutscene.ShowDialog() == DialogResult.OK)
            {
                StreamWriter writer = File.CreateText(mSfdCutscene.FileName);
                writer.Write(mTxtBoxEditor.Text);
                writer.Close();

                SetCutsceneFile(mSfdCutscene.FileName);
            }
        }

        private void mMainMenuFileOpenExe_Click(object sender, EventArgs e)
        {
            mOfdCutsceneExe.FileName = mCutsceneFile != null ? Path.ChangeExtension(mCutsceneFile.FullName, "cexe") : "";
            mOfdCutsceneExe.InitialDirectory = !string.IsNullOrEmpty(mOfdCutsceneExe.FileName) ? Path.GetDirectoryName(mOfdCutsceneExe.FileName) : ".";
            if (mOfdCutsceneExe.ShowDialog() == DialogResult.OK)
            {
                Executable cutsceneExe = Executable.LoadFromFile(mOfdCutsceneExe.FileName);
                if (cutsceneExe != null)
                {
                    SetCutsceneFile(null);

                    // Stop the current script                    
                    mScript.Stop();
                    UpdatePlaybackControl();

                    // Use the loaded executable
                    mScript = Script.CreateScript(mScene, cutsceneExe);

                    // Attempt to disassemble the loaded executable
                    string cutscene = Disassembler.Disassemble(cutsceneExe);
                    if (!string.IsNullOrEmpty(cutscene))
                        mTxtBoxEditor.Text = cutscene;
                    else
                        mTxtBoxEditor.Text = "Unable to disassemble executable.";
                }
                else
                {
                    SetCutsceneFile(null);

                    mTxtBoxEditor.Text = "Unable to load executable.";
                }
            }
        }

        private void mMainMenuFileSaveExeAs_Click(object sender, EventArgs e)
        {
            Compiler.Result compilerResult = CompileCutscene();
            if (!compilerResult.HasErrors)
            {
                mSfdCutsceneExe.FileName = mCutsceneFile != null ? Path.ChangeExtension(mCutsceneFile.FullName, "cexe") : "";
                mSfdCutsceneExe.InitialDirectory = !string.IsNullOrEmpty(mSfdCutsceneExe.FileName) ? Path.GetDirectoryName(mSfdCutsceneExe.FileName) : ".";
                if (mSfdCutsceneExe.ShowDialog() == DialogResult.OK)
                {
                    compilerResult.Executable.Save(mSfdCutsceneExe.FileName);
                }
            }
        }

        private void mMainMenuFileExit_Click(object sender, EventArgs e)
        {
            Close();
        }
        #endregion
    }
}
