﻿using System;
using System.IO;
using CutsceneCompiler;

namespace CutsceneEditor
{
    public static class Interpreter
    {
        public static bool ExecuteOp(Routine.ExecutionContext ctx)
        {
            ushort cur_pc = ctx[Routine.Registers.PC];
            ushort new_pc = cur_pc;

            OpCode opCode = (OpCode)ctx.Executable[new_pc++];
            switch (opCode)
            {
                case OpCode.RETURN:
                    cur_pc = new_pc;
                    break;

                case OpCode.YIELD:
                    ctx.SetStatus(Routine.StatusFlags.Yield);
                    cur_pc = new_pc;
                    break;

                case OpCode.SETPC:
                    {
                        cur_pc = FetchAddress(ctx, ref new_pc);
                    } break;

                case OpCode.StartRoutine:
                    {
                        StartRoutine(ctx, ref new_pc);
                        cur_pc = new_pc;
                    } break;

                case OpCode.Wait:
                    {
                        if (Wait(ctx, ref new_pc))
                            cur_pc = new_pc;

                    } break;

                case OpCode.CreateActor:
                    {
                        CreateActor(ctx, ref new_pc);
                        cur_pc = new_pc;

                    } break;

                case OpCode.MoveActor:
                    {
                        MoveActor(ctx, ref new_pc);
                        cur_pc = new_pc;

                    } break;

                case OpCode.WalkActor:
                    {
                        if (WalkActor(ctx, ref new_pc))
                            cur_pc = new_pc;

                    } break;

                case OpCode.StartAnimActor:
                    {
                        StartAnimActor(ctx, ref new_pc);
                        cur_pc = new_pc;

                    } break;

                case OpCode.ShowActorText:
                    {
                        if (ShowText(ctx, ref new_pc, true))
                            cur_pc = new_pc;

                    } break;

                case OpCode.ShowText:
                    {
                        if (ShowText(ctx, ref new_pc, false))
                            cur_pc = new_pc;

                    } break;
            }

            ctx[Routine.Registers.PC] = cur_pc;

            return opCode != OpCode.RETURN;
        }

        private static ushort FetchLiteral(Routine.ExecutionContext ctx, ref ushort pc)
        {
            ushort value = ctx.Executable[pc++];
            if ((value & 0x0080) != 0)
            {
                if ((value & 0x00c0) == 0x0080)
                {
                    int high = value & 0x003f;
                    int low = ctx.Executable[pc++];
                    value = (ushort)((high << 8) + low);
                }
                else
                {
                    int high = ctx.Executable[pc++];
                    int low = ctx.Executable[pc++];
                    value = (ushort)((high << 8) + low);
                }
            }

            return value;
        }

        private static ushort FetchDuration(Routine.ExecutionContext ctx, ref ushort pc)
        {
            return FetchLiteral(ctx, ref pc);
        }

        private static ushort FetchAddress(Routine.ExecutionContext ctx, ref ushort pc)
        {
            int high = ctx.Executable[pc++];
            int low = ctx.Executable[pc++];
            return (ushort)((high << 8) + low);
        }

        private static Actor FetchActor(Routine.ExecutionContext ctx, ref ushort pc)
        {
            ushort actorId = FetchLiteral(ctx, ref pc);
            string actorName = ctx.Executable.GetString(actorId);

            Actor actor = ctx.Scene.GetActor(actorName);
            if (actor == null)
                throw new InvalidDataException("Unknown actor: " + actorName);

            return actor;
        }

        private static void StartRoutine(Routine.ExecutionContext ctx, ref ushort pc)
        {
            ushort address = FetchAddress(ctx, ref pc);
            ctx.Script.StartRoutine(address);
        }

        private static bool Wait(Routine.ExecutionContext ctx, ref ushort pc)
        {
            ushort waitTimeMs = FetchDuration(ctx, ref pc);

            if (ctx[Routine.Registers.D] == 0xbeef)
            {
                ushort timeElapsedMs = (ushort)(1000 / ctx.Script.ClockSpeed);
                ctx[Routine.Registers.A] += timeElapsedMs;

                if (ctx[Routine.Registers.A] >= waitTimeMs)
                {
                    ctx[Routine.Registers.A] = 0;
                    ctx[Routine.Registers.D] = 0;
                    return true;
                }
            }
            else
            {
                ctx[Routine.Registers.A] = 0;
                ctx[Routine.Registers.D] = 0xbeef;
            }

            ctx.SetStatus(Routine.StatusFlags.Yield);
            return false;
        }

        private static void CreateActor(Routine.ExecutionContext ctx, ref ushort pc)
        {
            ushort actorId = FetchLiteral(ctx, ref pc);
            string actorName = ctx.Executable.GetString(actorId);

            ushort actorTypeId = FetchLiteral(ctx, ref pc);
            string actorType = ctx.Executable.GetString(actorTypeId);

            Actor newActor = Actor.CreateActor(actorName, actorType);

            ctx.Script.Scene.AddActor(newActor);
        }

        private static void MoveActor(Routine.ExecutionContext ctx, ref ushort pc)
        {
            Actor actor = FetchActor(ctx, ref pc);
            actor.X = FetchLiteral(ctx, ref pc);
            actor.Y = FetchLiteral(ctx, ref pc);
        }

        private static bool WalkActor(Routine.ExecutionContext ctx, ref ushort pc)
        {
            Actor actor = FetchActor(ctx, ref pc);

            int targetX = FetchLiteral(ctx, ref pc);
            int targetY = FetchLiteral(ctx, ref pc);

            bool moveX = actor.X != targetX || actor.U != 0;
            bool moveY = actor.Y != targetY || actor.V != 0;
            if (!moveX && !moveY)
            {
                actor.StartAnimation("stand_front");
                actor.DU = 0;
                actor.DV = 0;

                ctx.UnsetStatus(Routine.StatusFlags.Yield);
                return true;
            }

            bool walkRight = targetX > actor.X;
            string walkAnimationName = walkRight ? "walk_right" : "walk_left";

            bool startAnim = actor.CurrentAnimation == null || actor.CurrentAnimation.Name != walkAnimationName;
            if (startAnim)
                actor.StartAnimation(walkAnimationName);

            actor.DU = moveX ? (walkRight ? 1 : -1) : 0;
            actor.DV = moveY ? (targetY > actor.Y ? 1 : -1) : 0;

            ctx.SetStatus(Routine.StatusFlags.Yield);
            return false;
        }

        private static void StartAnimActor(Routine.ExecutionContext ctx, ref ushort pc)
        {
            Actor actor = FetchActor(ctx, ref pc);

            ushort animNameId = FetchLiteral(ctx, ref pc);
            string animName = ctx.Executable.GetString(animNameId);

            actor.StartAnimation(animName);
        }

        private static bool ShowText(Routine.ExecutionContext ctx, ref ushort pc, bool actorText)
        {
            Actor actor = actorText ? FetchActor(ctx, ref pc) : null;

            ushort stringId = FetchLiteral(ctx, ref pc);
            string textValue = ctx.Executable.GetString(stringId);

            ushort textID = ctx[Routine.Registers.B];
            Text text = ctx.Script.Scene.GetText(textID);

            bool isDone = Wait(ctx, ref pc);
            if (isDone)
            {
                ctx.Script.Scene.RemoveText(text);
            }
            else
            {
                if (text == null)
                {
                    text = new Text(textValue, actor);
                    ctx.Script.Scene.AddText(text);
                    ctx[Routine.Registers.B] = text.ID;
                }
            }

            return isDone;
        }
    }
}
