﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Timers;
using CutsceneCompiler;

namespace CutsceneEditor
{
    public class Script
    {
        private enum ExecutionState
        {
            Stopped,
            Running,
            Paused,
        }

        public Scene Scene { get; private set; }

        public Executable Executable { get; private set; }

        private List<Routine> mRoutines;
        public IEnumerable<Routine> Routines { get { return mRoutines; } }

        public ushort ClockSpeed { get; private set; }
        private Timer mClock;

        private ExecutionState mState;
        private bool mIsDone;

        public bool IsRunning { get { return mState == ExecutionState.Running; } }
        public bool IsPaused { get { return mState == ExecutionState.Paused; } }
        public bool IsDone { get { return mIsDone && mState == ExecutionState.Stopped; } }

        private uint mRunTimeMs;
        public uint RunTimeMs { get { return mRunTimeMs; } }

        protected Script(Scene scene, Executable executable)
        {
            Scene = scene;

            Executable = executable;

            mRoutines = new List<Routine>();

            ClockSpeed = 100; // 100 Hz

            mClock = new Timer();
            mClock.Interval = 1000 / ClockSpeed; 
            mClock.Elapsed += new ElapsedEventHandler(mClock_Elapsed);

            mState = ExecutionState.Stopped;
            mIsDone = false;

            mRunTimeMs = 0;
        }

        public void Start(bool manualTick)
        {
            if (mState == ExecutionState.Running)
                return;

            if (mState == ExecutionState.Stopped)
            {
                Scene.Reset();
                StartRoutine(0);
                mIsDone = false;
                mRunTimeMs = 0;
            }

            mState = ExecutionState.Running;

            if (!manualTick)
                mClock.Start();
        }

        public void Stop()
        {
            if (mState == ExecutionState.Stopped)
                return;

            mClock.Stop();

            mState = ExecutionState.Stopped;
            mRunTimeMs = 0;

            mRoutines.Clear();
        }

        public void Pause()
        {
            if (mState != ExecutionState.Running)
                return;

            mClock.Stop();

            mState = ExecutionState.Paused;
        }

        public Routine CreateRoutine(ushort address)
        {
            return new Routine(this, address);
        }

        public void StartRoutine(ushort address)
        {
            Routine routine = CreateRoutine(address);
            if (routine != null)
                mRoutines.Add(routine);
        }

        public Routine GetRoutine(string name)
        {
            foreach (Routine routine in mRoutines)
                if (routine.Name == name)
                    return routine;

            return null;
        }

        private void mClock_Elapsed(object sender, ElapsedEventArgs e)
        {
            mClock.Stop();

            DateTime startTime = DateTime.Now;

            Tick();

            // Adjust the wait time based on how long the current tick was
            TimeSpan duration = DateTime.Now.Subtract(startTime);
            int elapsedTimeMs = Math.Min((int)duration.TotalMilliseconds, 1000);
            int nextInterval = (1000 - elapsedTimeMs) / ClockSpeed;
            mClock.Interval = 1000.0 / (double)Math.Max(nextInterval, ClockSpeed);

            mClock.Start();
        }

        public void Tick()
        {
            if (mState == ExecutionState.Running)
            {
                Execute();

                if (mRoutines.Count <= 0)
                {
                    mState = ExecutionState.Stopped;
                    mIsDone = true;
                }

                mRunTimeMs += (uint)(1000 / ClockSpeed);
            }
        }

        private void Execute()
        {
            int numRoutines = mRoutines.Count;
            for (int i = 0; i < numRoutines; i++)
                mRoutines[i].Execute();

            for (int i = mRoutines.Count - 1; i >= 0; i--)
            {
                if (mRoutines[i].IsDone)
                    mRoutines.RemoveAt(i);
            }
        }

        public override string ToString()
        {
            return Executable.Name + " (" + mState.ToString() + ")";
        }

        public static Script CreateScript(Scene scene, Executable executable)
        {
            return new Script(scene, executable);
        }

        public static Script CreateTestScript(Scene scene, int testId)
        {
            if (testId == 1)
            {
                ExecutableWriter exeWriter = new ExecutableWriter("Robot meeting!");

                // Define the required strings
                exeWriter.AddString("main");
                exeWriter.AddString("redbot_main");
                exeWriter.AddString("bluebot_main");
                exeWriter.AddString("Redbot");
                exeWriter.AddString("Bluebot");
                exeWriter.AddString("stand_back");
                exeWriter.AddString("Robot meeting!");

                // Create the byte code for the co-routines
                ExecutableWriter redbotWriter = new ExecutableWriter("Redbot");
                ExecutableWriter bluebotWriter = new ExecutableWriter("Bluebot");

                // BEGIN redbot_main
                redbotWriter.AddLiteral(1); // String: redbot_main
                // Walk Redbot to the right
                redbotWriter.AddOpCode(OpCode.WalkActor);
                redbotWriter.AddLiteral(3);
                redbotWriter.AddLiteral(3);
                redbotWriter.AddLiteral(0);
                // Wait for a second
                redbotWriter.AddOpCode(OpCode.Wait);
                redbotWriter.AddDuration(1, 0);
                // Look into the distance
                redbotWriter.AddOpCode(OpCode.StartAnimActor);
                redbotWriter.AddLiteral(3);
                redbotWriter.AddLiteral(5); // String: stand_back
                // Wait a bit longer
                redbotWriter.AddOpCode(OpCode.Wait);
                redbotWriter.AddDuration(1, 0);
                // Walk Redbot to center
                redbotWriter.AddOpCode(OpCode.WalkActor);
                redbotWriter.AddLiteral(3);
                redbotWriter.AddLiteral(7);
                redbotWriter.AddLiteral(0);
                // END redbot_main
                redbotWriter.AddOpCode(OpCode.RETURN);

                // BEGIN bluebot_main
                bluebotWriter.AddLiteral(2); // String: bluebot_main
                // Walk Bluebot to center
                bluebotWriter.AddOpCode(OpCode.WalkActor);
                bluebotWriter.AddLiteral(4);
                bluebotWriter.AddLiteral(8);
                bluebotWriter.AddLiteral(0);
                // END redbot_main
                bluebotWriter.AddOpCode(OpCode.RETURN);

                // Create the program byte code
                // BEGIN main
                exeWriter.AddLiteral(0); // String: main
                exeWriter.AddOpCode(OpCode.SETPC); // Jump over sub-routine code
                int address_offset = exeWriter.DataLength + redbotWriter.DataLength + bluebotWriter.DataLength;
                exeWriter.AddAddress((ushort)(address_offset + 2)); // +2 because offsets (and addresses) are always two bytes

                // BEGIN redbot_main
                ushort redbot_main_address = exeWriter.DataLength;
                exeWriter.AddExecutable(redbotWriter);

                // BEGIN bluebot_main
                ushort bluebot_main_address = exeWriter.DataLength;
                exeWriter.AddExecutable(bluebotWriter);

                // BEGIN main
                // Create Redbot
                exeWriter.AddOpCode(OpCode.CreateActor);
                exeWriter.AddLiteral(3); // String: Redbot
                exeWriter.AddLiteral(3); // String: Redbot
                exeWriter.AddOpCode(OpCode.MoveActor);
                exeWriter.AddLiteral(3); // String: Redbot
                exeWriter.AddLiteral(0); // Location x
                exeWriter.AddLiteral(0); // Location y
                // Create Redbot routine
                exeWriter.AddOpCode(OpCode.StartRoutine);
                exeWriter.AddAddress(redbot_main_address);
                // Create Bluebot
                exeWriter.AddOpCode(OpCode.CreateActor);
                exeWriter.AddLiteral(4); // String: Bluebot
                exeWriter.AddLiteral(4); // String: Bluebot
                exeWriter.AddOpCode(OpCode.MoveActor);
                exeWriter.AddLiteral(4); // String: Bluebot
                exeWriter.AddLiteral(15); // Location x
                exeWriter.AddLiteral(0); // Location y
                // Create Bluebot routine
                exeWriter.AddOpCode(OpCode.StartRoutine);
                exeWriter.AddAddress(bluebot_main_address);
                // Create text
                exeWriter.AddOpCode(OpCode.ShowText);
                exeWriter.AddLiteral(6); // String: Robot meeting!
                exeWriter.AddDuration(2, 0);
                // END main
                exeWriter.AddOpCode(OpCode.RETURN);

                return new Script(scene, exeWriter.CreateExecutable());
            }
            else
            {
                ExecutableWriter exeWriter = new ExecutableWriter("Walking robots");

                // Define the required strings
                exeWriter.AddString("main");
                exeWriter.AddString("Redbot");
                exeWriter.AddString("Bluebot");

                // Create the program byte code
                // BEGIN main
                exeWriter.AddLiteral(0); // String: main
                // Create Redbot
                exeWriter.AddOpCode(OpCode.CreateActor);
                exeWriter.AddLiteral(1); // String: Redbot
                exeWriter.AddLiteral(1); // String: Redbot
                exeWriter.AddOpCode(OpCode.MoveActor);
                exeWriter.AddLiteral(1); // String: Redbot
                exeWriter.AddLiteral(0); // Location x
                exeWriter.AddLiteral(0); // Location y
                // Create Bluebot
                exeWriter.AddOpCode(OpCode.CreateActor);
                exeWriter.AddLiteral(2); // String: Bluebot
                exeWriter.AddLiteral(2); // String: Bluebot
                exeWriter.AddOpCode(OpCode.MoveActor);
                exeWriter.AddLiteral(2); // String: Bluebot
                exeWriter.AddLiteral(15); // Location x
                exeWriter.AddLiteral(0); // Location y
                // Yield to other routines
                exeWriter.AddOpCode(OpCode.YIELD);
                // Walk Redbot to center
                exeWriter.AddOpCode(OpCode.WalkActor);
                exeWriter.AddLiteral(1);
                exeWriter.AddLiteral(7);
                exeWriter.AddLiteral(0);
                // Walk Bluebot to center
                exeWriter.AddOpCode(OpCode.WalkActor);
                exeWriter.AddLiteral(2);
                exeWriter.AddLiteral(8);
                exeWriter.AddLiteral(0);
                // END main
                exeWriter.AddOpCode(OpCode.RETURN);

                return new Script(scene, exeWriter.CreateExecutable());
            }
        }
    }
}
